/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#ifndef HPP_LOS_LINK_H
#define HPP_LOS_LINK_H

#include <cstdio>
#include "MatrixAbstractLayer/MatrixAbstractLayer.h"
class ChppGikStandingRobot;
class CjrlJoint;
class CjrlHumanoidDynamicRobot;

//#include "hppGik/robot/hppGikStandingRobot.h"
//#include "hpplocalstepper/hppLosGeometry.hh"

/**
Data computed by ChppLosLink objects and required for computations of ChppLosFeetAntiCollisionTask and ChppLosFeetAreaTask objects
 */
class ChppLosJacobianPack
{
    public:
        double LinRx;
        double LinRy;
        double LinRth;
        double ULinRx;
        double ULinRy;
        double DLinRx;
        double DLinRy;
        matrixNxP jacLinR;
        matrixNxP jacLUinR;
        matrixNxP jacLDinR;
        vectorN   jacLYawinR;
        
        double RinLx;
        double RinLy;
        double RinLth;
        double URinLx;
        double URinLy;
        double DRinLx;
        double DRinLy;
        matrixNxP jacRinL;
        matrixNxP jacRUinL;
        matrixNxP jacRDinL;
        vectorN   jacRYawinL;
};

/**
A rectangular 2D shape expressed by the coordinates of its 4 corners
 */
class ChppLosFootprintShape
{
    public:
        double xUpRight,yUpRight;
        double xDownRight,yDownRight;
        double xDownLeft,yDownLeft;
        double xUpLeft,yUpLeft;
};

/**
Definition of a configuration region for left footprint with respect to right footprint
 */
class ChppLosFootprintArea
{
    public:
        double maxX,minX,maxY,minY,maxTheta,minTheta;
};

/**
A footprint defined by its shape and position in world coordinates
 */
class ChppLosFootprint
{
    public:
        double x,y,th;
        double UXx,UYx;
        double UXy,UYy;
        void print() const
        {
            std::cout << x<<" "<<y<<" "<<th<<" "<< std::endl;
        }
};



class ChppLosLink;
/**
The base component of a ChppLosCentipede object. It represents a deformable support polygon. The deformation is controlled via a configuration vector that can be set through method applyConfiguration() and retrieved via getConfiguration(). An instance holds its rank in the configuration vector of the owning ChppLosCentipede object, as well as its number of degrees of freedom and its parent link.
\ingroup parts
 */
class ChppLosLink
{
    public:
        /**
        \brief Constructor
         */
        ChppLosLink ( unsigned int inNumberDof, bool rootIsRight, const ChppLosFootprintArea& leftStepArea,  const ChppLosFootprintArea& righttStepArea, const ChppLosFootprintShape& inLeftFootShape, const ChppLosFootprintShape& inRightFootShape );

        /**
        \brief Get the data for Feet task computations
         */
        const ChppLosJacobianPack& jacobianPack();

        /**
        \brief Destructor
         */
        virtual ~ChppLosLink() {};

        /**
        \brief Set the rank in owning ChppLosCentipede object
         */
        void rank ( unsigned int rank );
        /**
        \brief Get the rank in owning ChppLosCentipede object
         */
        unsigned int rank() const;
        /**
        \brief Get the number of degrees of freedom
         */
        unsigned int numberDof() const;

        /**
        \brief Set the parent link according to organization in owning ChppLosCentipede object
         */
        void parent ( ChppLosLink* inP );
        /**
        \brief Get the parent link according to organization in owning ChppLosCentipede object
         */
        ChppLosLink* parent() const;

        /**
        \brief Get the left step area
         */
        const ChppLosFootprintArea& leftStepArea() const;
        /**
        \brief Get the right step area
         */
        const ChppLosFootprintArea& rightStepArea() const;
        
        /**
        \brief Get the left footprint shape
         */
        const ChppLosFootprintShape& leftShapeArea() const;
        /**
        \brief Get the right footprint shape
         */
        const ChppLosFootprintShape& rightShapeArea() const;
        
        /**
        \brief Tell is the root footprint is the right footprint
         */
        bool rootIsRight() const;
        
        /**
        \brief Get the left footprint
         */
        const ChppLosFootprint& leftprint() const;
        /**
        \brief Get the right footprint
         */
        const ChppLosFootprint& rightprint() const;
        /**
        \brief Get the front footprint, which may be connected to the next link
         */
        const ChppLosFootprint& frontprint() const;
        /**
        \brief Get the root footprint, which may be connected to the previous link
         */
        const ChppLosFootprint& rootprint() const;
        /**
        \brief Tell whether root footprint is right
         */
        bool isRootprintRight() const;
        
        /**
        \brief Return current configuration
         */
        const vectorN& currentConfiguration();

        /**
        \brief Apply configuration from given centipede configuration vector
         */
        virtual bool applyConfiguration ( const ChppLosFootprint& leftprint, const ChppLosFootprint& rightprint, const vectorN& inConfiguration, bool computeRelativeFeetjacobians = true ) = 0;

        /**
        \brief Get the full jacobian of a frame attached to the front foot and whose center is at given 3d position
         */
        virtual bool getJacobian ( const vector3d& inPoint, matrixNxP& outjacobian, unsigned int inStart ) = 0;
        
        /**
        \brief Get the position jacobian of a frame attached to the front foot and whose center is at given 3d position
         */
        virtual bool getPositionJacobian ( const vector3d& inPoint, matrixNxP& outjacobian, unsigned int inStart ) = 0;
        
        /**
        \brief Get the orientation jacobian of a frame attached to the front foot
         */
        virtual bool getOrientationJacobian (matrixNxP& outjacobian, unsigned int inStart ) = 0;
        
        /**
        \brief Write the current configuration of this link on the given vector at this link's rank
         */
        bool getConfiguration(vectorN& outVec);
        /**
        \brief Get the upper velocity bounds of this link's parameters
         */
        bool getUpperVelocityBounds(vectorN& outVec);
        /**
        \brief Get the lower velocity bounds of this link's parameters
         */
        bool getLowerVelocityBounds(vectorN& outVec);

    protected:
        void computeRelativeCoordinates();
        /**
        \brief Compute the data required for feet related constraints
         */
        virtual void computeJacobianPack () = 0;
        bool attRootIsRight;
        unsigned int attRank;
        unsigned int attNumberDof;
        ChppLosFootprint attLeftprint,attRightprint;
        ChppLosFootprint *attRootprint, *attFrontprint;
        ChppLosJacobianPack attJP;
        ChppLosLink* attParent;
        vectorN attConfiguration,attUb,attLb;
        ChppLosFootprintArea attLeftArea, attRightArea;
        ChppLosFootprintShape attLeftShape,attRightShape;
};

/**
A link made of the relative position of left footprint with respect to right footprint. The dimension is 3 (deltaX,deltaY,deltaTheta).
\ingroup parts
 */
class ChppLosSupportPolygonLink : public ChppLosLink
{
    public:
        /**
        \brief Constructor
         */
        ChppLosSupportPolygonLink ( bool rootIsRight, double inDX, double inDY, double inDth, const ChppLosFootprintArea& leftStepArea,  const ChppLosFootprintArea& righttStepArea, const ChppLosFootprintShape& inLeftFootShape, const ChppLosFootprintShape& inRightFootShape );
        /**
        \brief Inherited
         */
        bool applyConfiguration ( const ChppLosFootprint& leftprint, const ChppLosFootprint& rightprint, const vectorN& inConfiguration,bool computeRelativeFeetjacobians = true );
        /**
        \brief Inherited
         */
        bool getJacobian ( const vector3d& inPoint, matrixNxP& outjacobian, unsigned int inStart );
        /**
        \brief Inherited
         */
        bool getPositionJacobian ( const vector3d& inPoint, matrixNxP& outjacobian, unsigned int inStart );
        /**
        \brief Inherited
         */
        bool getOrientationJacobian (matrixNxP& outjacobian, unsigned int inStart );
    private:
        /**
        \brief Inherited
         */
        void computeJacobianPack ( );
};

/**
A link made of a CjrlHumanoidDynamicRobot object whose feet are flat on the ground and connect the previous and next links. The number od degrees of freedom is CjrlHumanoidDynamicRobot::numberDof()-6.
\ingroup parts
 */
class ChppLosRobotLink : public ChppLosLink
{
    public:
        /**
        \brief Constructor
         */
        ChppLosRobotLink ( ChppGikStandingRobot& inRobot, bool rootIsRight, const ChppLosFootprintArea& leftStepArea,  const ChppLosFootprintArea& righttStepArea, const ChppLosFootprintShape& inLeftFootShape, const ChppLosFootprintShape& inRightFootShape );
        /**
        \brief Get the associated standing robot
         */
        ChppGikStandingRobot& standingRobot();

        /**
        \brief  Change the root foot print.
         */
        void rootIsRight(bool newRootIsRight);
        
        /**
        \brief Set the transformation matrix defining the position and orientation of the robot root foot with respect to its footprint
         */
        void setFootFreeFlyerWrtFootprint(const matrix4d& inM);
                
        /**
        \brief Get the foot joint corresponding to the root footprint
         */
        CjrlJoint* rootJoint();
        /**
        \brief Get the foot joint corresponding to the front footprint
         */
        CjrlJoint* frontJoint();

        /**
        \brief Inherited
         */
        bool applyConfiguration ( const ChppLosFootprint& leftprint, const ChppLosFootprint& rightprint, const vectorN& inConfiguration,bool computeRelativeFeetjacobians = true );

        /**
        \brief Inherited
         */
        bool getJacobian ( const vector3d& inPoint, matrixNxP& outjacobian, unsigned int inStart );
        /**
        \brief Inherited
         */
        bool getPositionJacobian ( const vector3d& inPoint, matrixNxP& outjacobian, unsigned int inStart );
        /**
        \brief Inherited
         */
        bool getOrientationJacobian (matrixNxP& outjacobian, unsigned int inStart );
        /**
        \brief Modify upper limit of dof at given index
         */
        bool upperDofLimit (unsigned int index, double ul );
        /**
        \brief Modify lower limit of dof at given index
         */
        bool lowerDofLimit (unsigned int index, double ll );
        
        /**
        \update velocity bounds based on current configuration
        */
        void updateVelocityBounds();

    private:
        /**
        \brief Inherited
         */
        void computeJacobianPack ( );

        ChppGikStandingRobot* attStandingRobot;

        matrixNxP jac;
        matrix4d m,attPrint2Root;
        vectorN robotConfiguration;

        CjrlJoint* attFrontJoint;
        CjrlJoint* attRootJoint;
        matrix4d H0,Hif,InvHf,Hf,mroot;
        vector3d tmpv3d;
        matrix3d tmpr3d;
        double attLeftAnkleHeight,attRightAnkleHeight;
        vectorN attUL,attLL,attMV,attDs,att_inv_Di_Ds,attDi;
};

#endif
