#include "hpplocalstepper/tasks/hppLosCollisionAvoider.hh"
#include "hpplocalstepper/hppLosLinkGeometry.hh"
#include "hpphik/hppHikLinearSystem.hh"
#include "hpplocalstepper/hppLosCentipede.hh"
#include "hpplocalstepper/hppLosLink.hh"
#include "hppGik/robot/hppGikStandingRobot.h"
#include "robotDynamics/jrlHumanoidDynamicRobot.h"
#include <string.h>
#include <time.h>
#include <sys/time.h>

#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

ChppLosCollisionAvoider::ChppLosCollisionAvoider ( ChppLosCentipede& inCentipede ) :ChppLosTask ( inCentipede )
{
    attConstraint->vector().resize ( 0 );
    attConstraint->matrix().resize ( 0, attCentipede->numberDof() );
    attValue.resize ( 0 );

    noutM.resize ( 1,3 );

    tempJac3dLong.resize ( 3, attCentipede->numberDof() );
    tempJac3dLong.clear();
}

ChppLosCollisionAvoider::~ChppLosCollisionAvoider()
{
}

void ChppLosCollisionAvoider::crossProd3 ( const vector3d& v1, const vector3d& v2, vector3d& retV )
{
    retV ( 0 ) = v1 ( 1 ) *v2 ( 2 ) - v1 ( 2 ) *v2 ( 1 );
    retV ( 1 ) = v1 ( 2 ) *v2 ( 0 ) - v1 ( 0 ) *v2 ( 2 );
    retV ( 2 ) = v1 ( 0 ) *v2 ( 1 ) - v1 ( 1 ) *v2 ( 0 );
}



bool ChppLosCollisionAvoider::setObstaclePair ( LosCentipedeLinkGeometry& inRobotOb, LosGeometry& inFixedOb, double inDi, double inDs, double inXi )
{
    LosCylinder *fc=0;
    LosHalfspace *fh = 0;
    LosRobotCylinder *rc=0;
    LosFootprint *fp=0;

    rc = dynamic_cast<LosRobotCylinder*> ( &inRobotOb );
    if ( !rc )
    {
        fp = dynamic_cast<LosFootprint*> ( &inRobotOb );
        if ( !fp )
        {
            std::cout << "ChppLosCollisionAvoider::setObstaclePair unrecognized inRobotOb subclass" <<std::endl;
            return false;
        }
    }

    fh = dynamic_cast<LosHalfspace*> ( &inFixedOb );
    if ( !fh )
    {
        fc = dynamic_cast<LosCylinder*> ( &inFixedOb );
        if ( !fc )
        {
            std::cout << "ChppLosCollisionAvoider::setObstaclePair unrecognized LosFixedObject subclass" <<std::endl;
            return false;
        }
    }
    if ( fc && fp )
    {
        std::cout << "ChppLosCollisionAvoider::setObstaclePair unhandled combination of LosCentipedeLinkGeometry and LosFixedObject subclasses" <<std::endl;
        return false;
    }

    storeCentipedeObject ( inRobotOb );


    ObstacleDamper damper;
    if ( rc )
        damper.lobject = damper.robject = rc;
    else
    {
        damper.robject = 0;
        damper.lobject = fp;
    }

    damper.fobject = &inFixedOb;
    damper.Di_Ds = inDi - inDs;

    PointPair PP;
    damper.pointpairs.push_back ( PP );
    damper.pointpairs.push_back ( PP );

    if ( rc )
        if ( fh )
        {
            attRCFH.push_back ( std::make_pair ( rc, fh ) );
            damper.Ds = rc->radius() + inDs;
            attRCFHDampers.push_back ( damper );
            attDampers.push_back ( & ( attRCFHDampers[attRCFHDampers.size()-1] ) );
        }
        else
        {
            damper.pointpairs.push_back ( PP );
            damper.Ds = rc->radius() + fc->radius() + inDs;
            attRCFCDampers.push_back ( damper );
            attRCFC.push_back ( std::make_pair ( rc, fc ) );
            attDampers.push_back ( & ( attRCFCDampers[attRCFCDampers.size()-1] ) );
        }
    else
    {
        attFPFH.push_back ( std::make_pair ( fp, fh ) );
        damper.Ds = inDs;
        attFPFHDampers.push_back ( damper );

    }
    return true;
}



void ChppLosCollisionAvoider::storeCentipedeObject ( LosCentipedeLinkGeometry& inRobotOb )
{
    bool found = false;
    unsigned int sz = attCentipedeObjects.size();
    for ( unsigned int i=0;i<sz;i++ )
        if ( attCentipedeObjects[i] == &inRobotOb )
        {
            found = true;
            break;
        }
    if ( !found )
        attCentipedeObjects.push_back ( &inRobotOb );
}

void ChppLosCollisionAvoider::setSelfPair ( LosRobotCylinder& inOb1, LosRobotCylinder& inOb2, double inDi, double inDs, double inXi )
{
    storeCentipedeObject ( inOb1 );
    storeCentipedeObject ( inOb2 );

    SelfDamper damper;
    damper.Ds = inOb1.radius() + inOb2.radius() + inDs;
    damper.Di = inOb1.radius() + inOb2.radius() + inDi;
    damper.Di_Ds = inDi - inDs;
    damper.Xi = inXi;

    if ( inOb1.length() >inOb2.length() )
    {
        damper.object1 = &inOb1;
        damper.object2 = &inOb2;
        attRCRC.push_back ( std::make_pair ( &inOb1, &inOb2 ) );
    }
    else
    {
        damper.object1 = &inOb2;
        damper.object2 = &inOb1;
        attRCRC.push_back ( std::make_pair ( &inOb2, &inOb1 ) );

    }
    attRCRCDampers.push_back ( damper );
}

#define dotprod(x,y) x[0]*y[0]+x[1]*y[1]+x[2]*y[2]
#define vecnorm(x) sqrt(dotprod(x,x))
#define EPS 1e-8

void ChppLosCollisionAvoider::getCylinderPointPairs ( const LosCylinder& inC1, const LosCylinder& inC2, std::vector< PointPair >& outPointPairs )
{
    outPointPairs.clear();

    const vector3d& u = inC1.absoluteAB();
    const vector3d& v = inC2.absoluteAB();
    const vector3d& w = inC2.absoluteA()-inC1.absoluteA();
    double a = inC1.length() *inC1.length();
    double b = dotprod ( u,v );
    double c = inC2.length() *inC2.length();
    double d = dotprod ( w,u );
    double e = dotprod ( w,v );
    vector3d diff;
    double lam1,lam2;
    double ac_b2 = a*c-b*b;
    if ( ac_b2 > EPS )
    {

        lam1 = ( d*c-b*e ) / ( ac_b2 );
        lam2 = ( b*d-e*a ) / ( ac_b2 );

        if ( lam1 >1 )
        {
            lam1 = 1;
            lam2 = ( b-e ) /c;
        }
        else if ( lam1 <0 )
        {
            lam1 = 0;
            lam2 = -e /c;
        }
        
        if (lam2 > 1)
        {
            lam2 = 1;
            if (d+b>a)
                lam1 = 1;
            else if (d+b<0)
                lam1 = 0;
            else
                lam1 = (d+b)/a;
        }
        else if (lam2 < 0)
        {
            lam2 = 0;
            if (d>a)
                lam1 = 1;
            else if (d<0)
                lam1 = 0;
            else
                lam1 = d/a;
        }
        PP.point1 = inC1.localA() + inC1.localAB() *lam1;
        PP.point2 = inC2.localA() + inC2.localAB() *lam2;
        PP.point1abs = inC1.absoluteA() + inC1.absoluteAB() *lam1;
        PP.point2abs = inC2.absoluteA() + inC2.absoluteAB() *lam2;

        diff = PP.point2abs-PP.point1abs;
        PP.distance = vecnorm ( diff );
        
        PP.normalp1p2 = diff/PP.distance;
        outPointPairs.push_back ( PP );
    }

    if ( d>a )
    {
        PP.point1 = inC1.localB(); PP.point1abs = inC1.absoluteB();
        PP.point2 = inC2.localA(); PP.point2abs = inC2.absoluteA();
    }
    else if ( d<0 )
    {
        PP.point1 = inC1.localA(); PP.point1abs = inC1.absoluteA();
        PP.point2 = inC2.localB(); PP.point2abs = inC2.absoluteB();
    }
    else
    {
        double Lam = d/a;
        PP.point1 = inC1.localA() + inC1.localAB() *Lam;
        PP.point1abs = inC1.absoluteA() + inC1.absoluteAB() *Lam;
        PP.point2 = inC2.localA(); PP.point2abs = inC2.absoluteA();
    }
    diff = PP.point2abs-PP.point1abs;
    PP.distance = vecnorm ( diff );
    PP.normalp1p2 = diff/PP.distance;

    double coordB2 = d+b;
    if ( coordB2>a )
    {
        PP.point1 = inC1.localB(); PP.point1abs = inC1.absoluteB();
        PP.point2 = inC2.localA(); PP.point2abs = inC2.absoluteA();
    }
    else if ( d<0 )
    {
        PP.point1 = inC1.localA(); PP.point1abs = inC1.absoluteA();
        PP.point2 = inC2.localB(); PP.point2abs = inC2.absoluteB();
    }
    else
    {
        double Lam = d/a;
        PP.point1 = inC1.localA() + inC1.localAB() *Lam;
        PP.point1abs = inC1.absoluteA() + inC1.absoluteAB() *Lam;
        PP.point2 = inC2.localA(); PP.point2abs = inC2.absoluteA();
    }
    diff = PP.point2abs-PP.point1abs;
    PP.distance = vecnorm ( diff );
    PP.normalp1p2 = diff/PP.distance;

    outPointPairs.push_back ( PP );
    outPointPairs.push_back ( PP );
}

void ChppLosCollisionAvoider::getCylinderHalfspacePointPairs ( const LosCylinder& inC, const LosHalfspace& inH, std::vector< PointPair >& outPointPairs )
{
    vector3d diff =  inC.absoluteA() - inH.borderPoint();
    double dist1 =  inH.outerNormal() [0]*diff[0] + inH.outerNormal() [1]*diff[1] + inH.outerNormal() [2]*diff[2];

    diff =  inC.absoluteB()- inH.borderPoint();
    double dist2 =  inH.outerNormal() [0]*diff[0] + inH.outerNormal() [1]*diff[1] + inH.outerNormal() [2]*diff[2];

    outPointPairs[0].distance = dist1;
    outPointPairs[0].point1abs = inC.absoluteA();

    outPointPairs[1].distance = dist2;
    outPointPairs[1].point1abs = inC.absoluteB();
}

void ChppLosCollisionAvoider::getFootprintHalfspacePointPair ( const std::vector<vector3d>& inVertices, const LosHalfspace& inH, std::vector< PointPair >& outPointPairs )
{
    const vector3d* vects[4];
    double dists[4];

    const vector3d& P = inH.borderPoint();
    const vector3d& N = inH.outerNormal();
    vector3d diff;
    double dist;
    for ( unsigned int i=0;i<4;i++ )
    {
        diff = inVertices[i]-P;
        dist = N[0]*diff[0] + N[1]*diff[1] + N[2]*diff[2];
        vects[i] = &inVertices[i];
        dists[i] = dist;
    }

    int j = 0;
    while ( j<3 )
    {
        bool no_change = true;
        const vector3d* swapVec;
        double swapVal;
        for ( unsigned int i=0;i<3;i++ )
            if ( dists[i] > dists[i+1] )
            {
                swapVec = vects[i+1];
                swapVal = dists[i+1];
                no_change = false;
                dists[i+1] = dists[i];
                vects[i+1] = vects[i];
                dists[i] = swapVal;
                vects[i] = swapVec;
            }
        if ( no_change )
            break;
        j++;
    }


    outPointPairs[0].distance = dists[0];
    outPointPairs[0].point1abs = *vects[0];

    outPointPairs[1].distance = dists[1];
    outPointPairs[1].point1abs = *vects[1];
}


void ChppLosCollisionAvoider::calcSelfConstraint ( SelfDamper& inDamper, unsigned int constraintRank )
{
    ChppLosRobotLink& rlink = inDamper.object1->robotlink();
    unsigned int startRank = rlink.rank();
    unsigned int cr = constraintRank;


    if ( jac1d.size2() !=rlink.numberDof() )
        jac1d.resize ( 1,rlink.numberDof() );
    if ( tempJac3d.size2() !=rlink.numberDof() )
        tempJac3d.resize ( 3,rlink.numberDof() );

    for ( unsigned int i=0;i<inDamper.pointpairs.size();i++ )
    {

        tempJac3d.clear();
        rlink.standingRobot().robot()->getPositionJacobian ( inDamper.object1->joint(), inDamper.object2->joint(), inDamper.pointpairs[i].point2, tempJac3d, 0, false );

        noutM ( 0,0 ) = inDamper.pointpairs[i].normalp1p2[0];
        noutM ( 0,1 ) = inDamper.pointpairs[i].normalp1p2[1];
        noutM ( 0,2 ) = inDamper.pointpairs[i].normalp1p2[2];

        noalias ( jac1d ) = prod ( -noutM,tempJac3d );
        memcpy ( ( &attConstraint->matrix().data() [cr*attCentipede->numberDof() +startRank] ),&jac1d.data() [0],rlink.numberDof() *sizeof ( double ) );

        attConstraint->vector() ( cr ) = inDamper.Xi*( inDamper.pointpairs[i].distance - inDamper.Ds )/inDamper.Di_Ds;

        cr++;
    }
}

void ChppLosCollisionAvoider::calcObstacleConstraint ( ObstacleDamper& inDamper, unsigned int constraintRank )
{
    //std::cout << "calculating! "<<std::endl;
    ChppLosRobotLink& rlink = inDamper.robject->robotlink();
    unsigned int startRank = rlink.rank();
    unsigned int cr = constraintRank;

    for ( unsigned int i=0;i<inDamper.pointpairs.size();i++ )
    {
        tempJac3dLong.clear();
        rlink.standingRobot().robot()->getPositionJacobian ( *rlink.rootJoint(), inDamper.robject->joint(), inDamper.pointpairs[i].point1, tempJac3dLong, startRank, false );
        ChppLosLink* plink = rlink.parent();
        while ( plink!=0 )
        {
            plink->getPositionJacobian ( inDamper.pointpairs[i].point1abs,tempJac3dLong,plink->rank() );
            plink = plink->parent();
        }

        noutM ( 0,0 ) = inDamper.pointpairs[i].normalp1p2[0];
        noutM ( 0,1 ) = inDamper.pointpairs[i].normalp1p2[1];
        noutM ( 0,2 ) = inDamper.pointpairs[i].normalp1p2[2];

        noalias ( subrange ( attConstraint->matrix(),cr,cr+1,0,attCentipede->numberDof() ) ) = prod ( -noutM,tempJac3dLong );

        attConstraint->vector() ( cr ) = ( inDamper.pointpairs[i].distance - inDamper.Ds ) / inDamper.Di_Ds ;

        cr++;
    }
}


void ChppLosCollisionAvoider::calcLinkConstraint ( ObstacleDamper& inDamper, unsigned int constraintRank )
{
    tempJac3dLong.clear();
    unsigned int cr = constraintRank;

    for ( unsigned int i=0; i<2; i++ )
    {
        ChppLosLink* plink = inDamper.lobject->link().parent();
        while ( plink!=0 )
        {
            plink->getPositionJacobian ( inDamper.pointpairs[i].point1abs,tempJac3dLong,plink->rank() );
            plink = plink->parent();
        }

        noutM ( 0,0 ) = inDamper.pointpairs[i].normalp1p2[0];
        noutM ( 0,1 ) = inDamper.pointpairs[i].normalp1p2[1];
        noutM ( 0,2 ) = inDamper.pointpairs[i].normalp1p2[2];

        noalias ( subrange ( attConstraint->matrix(),cr,cr+1,0,attCentipede->numberDof() ) ) = prod ( -noutM,tempJac3dLong );

        attConstraint->vector() ( cr ) = ( inDamper.pointpairs[i].distance - inDamper.Ds ) / inDamper.Di_Ds;

        cr++;
    }
}


void ChppLosCollisionAvoider::computeDifferentialSystem()
{
/*
       struct timeval *Tps, *Tpf;
       struct timezone *Tzp;

    
    Tps = ( struct timeval* ) malloc ( sizeof ( struct timeval ) );
    Tpf = ( struct timeval* ) malloc ( sizeof ( struct timeval ) );
    Tzp = 0;
    gettimeofday ( Tps, Tzp );
*/
    resize();

    for ( unsigned int i = 0; i< attCentipedeObjects.size(); i++ )
        attCentipedeObjects[i]->transform();


    unsigned int sizeConstraint = 0;

    for ( unsigned int i = 0;i<attRCRC.size();i++ )
    {
        getCylinderPointPairs ( *attRCRC[i].first, *attRCRC[i].second, attRCRCDampers[i].pointpairs );
        sizeConstraint += attRCRCDampers[i].pointpairs.size();
    }
    for ( unsigned int i = 0;i<attRCFC.size();i++ )
    {
        getCylinderPointPairs ( *attRCFC[i].first, *attRCFC[i].second, attRCFCDampers[i].pointpairs );
        sizeConstraint += attRCFCDampers[i].pointpairs.size();
    }
    for ( unsigned int i = 0;i<attRCFH.size();i++ )
    {
        getCylinderHalfspacePointPairs ( *attRCFH[i].first, *attRCFH[i].second, attRCFHDampers[i].pointpairs );
        sizeConstraint += attRCFHDampers[i].pointpairs.size();
    }
    for ( unsigned int i = 0;i<attFPFH.size();i++ )
    {
        getFootprintHalfspacePointPair ( attFPFH[i].first->vertices(), *attFPFH[i].second, attFPFHDampers[i].pointpairs );
        sizeConstraint += attFPFHDampers[i].pointpairs.size();
    }

    attConstraint->matrix().resize ( sizeConstraint, attCentipede->numberDof() );
    attConstraint->vector().resize ( sizeConstraint );
    attConstraint->matrix().clear();
    attConstraint->vector().clear();

    unsigned int curRank = 0;
    for ( unsigned int i = 0;i<attRCRCDampers.size();i++ )
    {
        calcSelfConstraint ( attRCRCDampers[i], curRank );
        curRank += attRCRCDampers[i].pointpairs.size();
    }

    for ( unsigned int i = 0;i<attRCFCDampers.size();i++ )
    {
        calcObstacleConstraint ( attRCFCDampers[i], curRank );
        curRank += attRCFCDampers[i].pointpairs.size();
    }

    for ( unsigned int i = 0;i<attRCFHDampers.size();i++ )
    {
        calcObstacleConstraint ( attRCFHDampers[i], curRank );
        curRank += attRCFHDampers[i].pointpairs.size();
    }

    for ( unsigned int i = 0;i<attFPFHDampers.size();i++ )
    {
        calcLinkConstraint ( attFPFHDampers[i], curRank );
        curRank += attFPFHDampers[i].pointpairs.size();
    }
/*
    gettimeofday ( Tpf, Tzp );
    printf ( "ComputeConstraint: %ld us\n", ( Tpf->tv_sec-Tps->tv_sec ) *1000000 + ( Tpf->tv_usec-Tps->tv_usec ) );
    
    free ( Tps );
    free ( Tpf );
    */
}


const vectorN& ChppLosCollisionAvoider::currentValue()
{
    attValue.clear();

    for ( unsigned int i = 0; i< attCentipedeObjects.size(); i++ )
        attCentipedeObjects[i]->transform();

    attDampers.clear();


    unsigned int sizeConstraint = 0;

    for ( unsigned int i = 0;i<attRCRC.size();i++ )
    {
        getCylinderPointPairs ( *attRCRC[i].first, *attRCRC[i].second, attRCRCDampers[i].pointpairs );
        sizeConstraint += attRCRCDampers[i].pointpairs.size();
        attDampers.push_back ( &attRCRCDampers[i] );
    }
    for ( unsigned int i = 0;i<attRCFC.size();i++ )
    {
        getCylinderPointPairs ( *attRCFC[i].first, *attRCFC[i].second, attRCFCDampers[i].pointpairs );
        sizeConstraint += attRCFCDampers[i].pointpairs.size();
        attDampers.push_back ( &attRCFCDampers[i] );
    }
    for ( unsigned int i = 0;i<attRCFH.size();i++ )
    {
        getCylinderHalfspacePointPairs ( *attRCFH[i].first, *attRCFH[i].second, attRCFHDampers[i].pointpairs );
        sizeConstraint += attRCFHDampers[i].pointpairs.size();
        attDampers.push_back ( &attRCFHDampers[i] );
    }
    for ( unsigned int i = 0;i<attFPFH.size();i++ )
    {
        getFootprintHalfspacePointPair ( attFPFH[i].first->vertices(), *attFPFH[i].second, attFPFHDampers[i].pointpairs );
        sizeConstraint += attFPFHDampers[i].pointpairs.size();
        attDampers.push_back ( &attFPFHDampers[i] );
    }


    attValue.resize ( sizeConstraint );

    unsigned int curRank = 0;
    for ( unsigned int i = 0;i<attDampers.size();i++ )
    {
        for ( unsigned int j=0;j<attDampers[i]->pointpairs.size();j++ )
        {
            //attDampers[i]->pointpairs[j].print(std::cout);
            //std::cout << attDampers[i]->pointpairs[j].distance << " " << attDampers[i]->Ds << std::endl;
            attValue ( curRank ) = ( attDampers[i]->pointpairs[j].distance < attDampers[i]->Ds ) ? ( attDampers[i]->pointpairs[j].distance - attDampers[i]->Ds ) : 0;
            curRank++;
        }
    }

    return attValue;
}



