#include "hpplocalstepper/tasks/hppLosComBoundingTask.hh"
#include "hpplocalstepper/hppLosCentipede.hh"
#include "hpphik/hppHikLinearSystem.hh"
#include "hpplocalstepper/hppLosLink.hh"
#include "hppGik/robot/hppGikStandingRobot.h"
#include "robotDynamics/jrlHumanoidDynamicRobot.h"
#include <string.h>

#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

ChppLosComBoundingTask::ChppLosComBoundingTask ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inRobotLink ) :ChppLosTask ( inCentipede )
{
    attRobotLink = &inRobotLink;
    attStandingRobot = &inRobotLink.standingRobot();

    attComXi = 0.02;
    attComDi = 0.12;
    attComDs = 0.095;

    jacCom.resize ( 3,attRobotLink->numberDof() );
    jacP.resize ( 3,attRobotLink->numberDof() );
    jac.resize ( 2,attRobotLink->numberDof() );
    jacO.resize ( 3,attRobotLink->numberDof() );
    jacF.resize ( 6,attRobotLink->numberDof() );
    jacVec.resize ( attRobotLink->numberDof() );
    
    vector3d anklePos;
    attStandingRobot->robot()->rightFoot()->getAnklePositionInLocalFrame(anklePos);
    attRightAnkleHeight = anklePos[2];
    
    attStandingRobot->robot()->leftFoot()->getAnklePositionInLocalFrame(anklePos);
    attLeftAnkleHeight = anklePos[2];

}

void ChppLosComBoundingTask::setMargins ( double inDi, double inDs, double inXi )
{
    attComXi = inXi;
    attComDs = inDs;
    attComDi = inDi;
}

void ChppLosComBoundingTask::computeDifferentialSystem()
{
    resize();
    CjrlHumanoidDynamicRobot* jrlRobot = attStandingRobot->robot();
    std::vector<const ChppGikLinkedVertex*> lvs;
    std::vector<const ChppGikLinkedVertex*> alvs;
    attStandingRobot->computeFeet2DConvexHull ( lvs );

    vector3d com = jrlRobot->positionCenterOfMass();
    //std::cout << com[0] << " " << com[1]  << std::endl;
    std::vector<ChppGik2DVertex> vs;
    std::vector<ChppGik2DVertex> nouts;
    std::vector<double> dists;
    std::vector<double> coefs;
    ChppGik2DVertex v;
    double value, coef;
    for ( unsigned int i=0;i<lvs.size();i++ )
    {
        value = ( com[0] - lvs[i]->world.x ) *lvs[i]->ntonext.x + ( com[1] - lvs[i]->world.y ) *lvs[i]->ntonext.y;//coordinate of projection of com on [lvs[i] lvs[i]->next]
        coef = value / lvs[i]->dtonext;

        if ( ( coef >= 1 ) || ( coef <= 0 ) )
            continue;

        v.x = value * lvs[i]->ntonext.x + lvs[i]->world.x;//world x position of projection
        v.y = value * lvs[i]->ntonext.y + lvs[i]->world.y;//world y position of projection
        value = sqrt ( pow ( ( com[0]-v.x ),2 ) +pow ( ( com[1]-v.y ),2 ) );

        if ( value <= attComDi )
        {
            alvs.push_back ( lvs[i] );
            vs.push_back ( v );
            dists.push_back ( value );
            coefs.push_back ( coef );
            nouts.push_back ( lvs[i]->nout );
        }
    }

    attConstraint->vector().resize ( alvs.size(),false );
    attConstraint->matrix().resize ( alvs.size(),attCentipede->numberDof(),false );
    attConstraint->matrix().clear();

    if ( alvs.size() != 0 )
    {
        for ( unsigned int i=0;i<alvs.size();i++ )
        {
            //compute and store inequality constraint values
            attConstraint->vector() ( i ) = attComXi* ( dists[i] - attComDs ) / ( attComDi- attComDs );
            //std::cout << "attConstraint->vector() (" << i <<  ") = " << attConstraint->vector() ( i ) << std::endl;
        }
        //std::cout << std::endl;

        unsigned int movingfoot = ( attRobotLink->frontJoint() ==jrlRobot->rightAnkle() ) ?1:0;
        unsigned int fixfoot = ( attRobotLink->rootJoint() == jrlRobot->rightAnkle() ) ?1:0;

        jrlRobot->getJacobianCenterOfMass ( *attRobotLink->rootJoint(), jacCom,0,false );
        jrlRobot->getOrientationJacobian ( *attRobotLink->rootJoint(),*attRobotLink->frontJoint(), jacO,0,false );

        unsigned int startRank = attRobotLink->rank();

        double footHeight;
        if (attRobotLink->frontJoint() == jrlRobot->leftAnkle())
            footHeight = attLeftAnkleHeight;
        else
            footHeight = attRightAnkleHeight;
        
        //std::cout << "-- start of com constraints " << std::endl;
        for ( unsigned int i=0;i<alvs.size();i++ )
        {
            jacP.clear();
            if ( ( alvs[i]->foot == fixfoot ) && ( alvs[i]->next->foot == fixfoot ) )
            {
                //std::cout << "Case 1 " << std::endl;
                jacVec = nouts[i].x*row ( jacCom,0 ) + nouts[i].y*row ( jacCom,1 );

                memcpy ( ( &attConstraint->matrix().data() [i*attCentipede->numberDof() +startRank] ),&jacVec.data() [0],attRobotLink->numberDof() *sizeof ( double ) );
            }

            else
                if ( ( alvs[i]->foot == movingfoot ) )
                {
                    if ( alvs[i]->next->foot == movingfoot )
                    {
                        //std::cout << "Case 2 " << std::endl;
                        vector3d localVec ( alvs[i]->local.x, alvs[i]->local.y, -footHeight );
                        jrlRobot->getPositionJacobian ( *attRobotLink->rootJoint(),*attRobotLink->frontJoint(), localVec, jacP,0,false );
                        jacVec = nouts[i].x* ( row ( jacCom,0 )-row ( jacP,0 ) ) + nouts[i].y* ( row ( jacCom,1 )-row ( jacP,1 ) );
                        double rotContrib = ( com[0] - alvs[i]->world.x ) *alvs[i]->ntonext.x + ( com[1] - alvs[i]->world.y ) *alvs[i]->ntonext.y;
                        jacVec += rotContrib*row ( jacO,2 );

                        memcpy ( ( &attConstraint->matrix().data() [i*attCentipede->numberDof() +startRank] ),&jacVec.data() [0],attRobotLink->numberDof() *sizeof ( double ) );
                    }
                    else // first vertex A is in moving foot, B is in fixed foot: M = B
                    {
                        //std::cout << "Case 3 " << std::endl;
                        vector3d localVec ( alvs[i]->local.x, alvs[i]->local.y, -footHeight );
                        jrlRobot->getPositionJacobian ( *attRobotLink->rootJoint(),*attRobotLink->frontJoint(), localVec, jacP,0,false );
                        row ( jac,0 ) = ( coefs[i] - 1 ) * row ( jacP,0 );
                        row ( jac,1 ) = ( coefs[i] - 1 ) * row ( jacP,1 );
                        jacVec = nouts[i].x* ( row ( jacCom,0 ) +row ( jac,0 ) ) + nouts[i].y* ( row ( jacCom,1 ) +row ( jac,1 ) );

                        memcpy ( ( &attConstraint->matrix().data() [i*attCentipede->numberDof() +startRank] ),&jacVec.data() [0],attRobotLink->numberDof() *sizeof ( double ) );
                    }
                }
                else // first vertex A is in fixed foot, B is in moving foot: M = A
                {
                    //std::cout << "Case 4 " << std::endl;
                    vector3d localVec ( alvs[i]->next->local.x, alvs[i]->next->local.y, -footHeight );
                    jrlRobot->getPositionJacobian ( *attRobotLink->rootJoint(),*attRobotLink->frontJoint(), localVec, jacP,0,false );
                    row ( jac,0 ) = -coefs[i] * row ( jacP,0 );
                    row ( jac,1 ) = -coefs[i] * row ( jacP,1 );
                    jacVec = nouts[i].x* ( row ( jacCom,0 ) +row ( jac,0 ) ) + nouts[i].y* ( row ( jacCom,1 ) +row ( jac,1 ) );

                    memcpy ( ( &attConstraint->matrix().data() [i*attCentipede->numberDof() +startRank] ),&jacVec.data() [0],attRobotLink->numberDof() *sizeof ( double ) );
                }
        }
    }
}



const vectorN& ChppLosComBoundingTask::currentValue()
{
    CjrlHumanoidDynamicRobot* jrlRobot = attStandingRobot->robot();
    std::vector<const ChppGikLinkedVertex*> lvs;
    attStandingRobot->computeFeet2DConvexHull ( lvs );
    vector3d com = jrlRobot->positionCenterOfMass();
    std::vector<double> dists;
    double value;
    for ( unsigned int i=0;i<lvs.size();i++ )
    {
        value = lvs[i]->nout.x * com[0] + lvs[i]->nout.y * com[1] - lvs[i]->b + attComDs;
        if (value > 0)
            dists.push_back (value);
    }

    attValue.resize ( dists.size() );
    for ( unsigned int i=0;i<dists.size();i++ )
        attValue ( i ) = dists[i];
    return attValue;
}

